from z3 import *
from functools import reduce
import itertools


class ExperimentVariable:
    """The Experiment Variable represents a user-defined factor or characteristic 
            that changes throughout the course of an experiment.

    name: user-defined name of the variable
    num_options: how many conditions we can assign to the variable
    possible_orders: list of possible assignment values
    """
     
    def __init__(self, name, num_options=0, options = []):
        # should we use argv or kwargs in place of a list?

        assert len(options) > 0 or num_options > 0

        num_options = len(options) if num_options == 0 else num_options

        assert num_options == len(options) or len(options) == 0
        assert num_options > 0

        self.name = name
        self.constraint = None
        self.n = num_options

        # experiment variables holds metadata, and information about 
        # possible assignments (ie. the conditions)
        self.conditions = []
        self.condition_map = dict()
        for i in range(num_options):
            # create a name for the condition if not defined
            attr = options[i] if len(options) > 0 else str(i)
            # instance of variable condition which represents a subcondition 
            # of this variable. Pass the current variable so the condition 
            # can refer back 
            self.conditions.append(attr)
            self.condition_map[attr] = i

    def get_condition(self, s):
        return self.conditions[self.condition_map[s]]

    def add_constraint(self, constraint):
        self.constraint = constraint

    def get_conditions(self):
        return self.conditions

    def __str__(self):
        return str(self.name)
    
    def __len__(self):
        return self.n
    
class DesignVariable:
    def __init__(self, var):
        assert isinstance(var, ExperimentVariable)
        self.var = var
        self.counterbalanced = False
        self.within_subjects = False

    def is_within_subjects(self, flag = True):
        self.within_subjects = flag

    def counterbalance(self, flag = True):
        self.counterbalanced = flag

class Replications(ExperimentVariable):
    def __init__(self, n):
        self.__init__("replications", n)

class MultiFactVariable(ExperimentVariable):
    def __init__(self, variables):
        self.n = reduce(lambda x, y: x*y, list(map(len, variables)))
        combinations = list(itertools.product(*[variable.conditions for variable in variables]))
        combinations = ["-".join(combination) for combination in combinations]

        super().__init__("factorial", self.n, options = combinations)
        all(isinstance(variable, ExperimentVariable) for variable in variables)
        self.variables = variables

    def get_variables(self):
        return self.variables


def multifact(variables):
    return MultiFactVariable(variables)

class BaseVariable(ExperimentVariable):
    def __init__(self):
        super().__init__("base", 1)


class VariableCondition:
    """The Variable represents a value that a user can assign to 
            a specific condition.

    NOTE: the ExperimentVariable() class is composed of VariableConditions
            - the user does not create these 

    name: user-defined name of the condition
    var: variable that this condition gets assigned to
             (the variable and condition can reference each-other)
    """
    def __init__(self, name, var):
        self.var = var # refer back to var object that constructed this option
        self.name = name

    def __str__(self):
        return self.name 
    
class UnitVariable:
    def __init__(self, name, levels):
        self.name = name
        self.levels = levels
    



# External libraries
from z3 import *                 # Requires: pip install z3-solver
import copy
import math
import pandas as pd

# Planet system imports
from planet.unit import Groups
from planet.variable import  multifact, ExperimentVariable
from planet.replications import Replications
from planet.design import Design, RandomPlan, Plans
from planet.constraint import (
    StartWith, Counterbalance, NoRepeat, InnerBlock, OuterBlock,
    SetRank, SetPosition, AbsoluteRank
)
from planet.designer import Designer
from planet.candl import combine_lists



def eval(designs):
    for design in designs:
        if design.groups is None:
            design._determine_num_plans()
        
        


def is_counterbalanced(d1, d2):
    return d1.counterbalanced or d2.counterbalanced

def nest_structure(d1, d2):
    constraints = []
    # Match all variables from the outer design within each block matrix
    
    for i in range(len(d2.variables)):
        constraints.append(InnerBlock(
            d2.variables[i],
            d1.get_width(),
            len(d1.groups),
            stride = [1, 1]
        ))

     # Match all variables from the inner design across every block

    for i in range(len(d1.variables)):
        constraints.append(OuterBlock(
            d1.variables[i],
            d1.get_width(),
            len(d1.groups),
            stride = [1, 1]
        ))

 
    return constraints

class ConstraintCopyVisitor:
    def __init__(self, design1, design2):
        self.design1 = design1
        self.design2 = design2
        self.width1 = design1.get_width()
        self.width2 = design2.get_width()
        self.total_groups = len(design1.groups) * len(design2.groups)
        self.total_conditions = self.width1 * self.width2
        self.constraints = []

    def visit(self, constraint, origin):
        method_name = f'visit_{constraint.__class__.__name__}'
        visitor = getattr(self, method_name, self.visit_default)
        return visitor(constraint, origin)

    def visit_Counterbalance(self, constraint, origin):
        if origin == 'design1':
            width = constraint.width or self.width1
            height = constraint.height or len(self.design1.groups)
            self.constraints.append(
                Counterbalance(
                    constraint.variable, width=width, height=height, stride=constraint.stride
                )
            )
        elif origin == 'design2':
            self.constraints.append(
                Counterbalance(
                    constraint.variable,
                    width=self.total_conditions,
                    height=self.total_groups,
                    stride=[len(self.design1.groups), self.width1 * constraint.stride[1]]
                )
            )

    def visit_NoRepeat(self, constraint, origin):
        if origin == 'design1':
            self.constraints.append(
                NoRepeat(constraint.variable, width=self.width1, stride=constraint.stride)
            )
        elif origin == 'design2':
            self.constraints.append(
                NoRepeat(
                    constraint.variable,
                    width=constraint.width * self.width1,
                    stride=constraint.stride * self.width1
                )
            )

    def visit_InnerBlock(self, constraint, origin):
        self.constraints.append(
            InnerBlock(
                constraint.variable,
                width=constraint.width * self.width1,
                height=constraint.height * len(self.design1.groups),
                stride=[1, 1]
            )
        )

    def visit_OuterBlock(self, constraint, origin):
        self.constraints.append(
            OuterBlock(
                constraint.variable,
                width=constraint.width * self.width1,
                height=constraint.height * len(self.design1.groups),
                stride=[1, 1]
            )
        )

    def visit_default(self, constraint, origin):
        self.constraints.append(copy.copy(constraint))

def accept(self, visitor, origin):
    return visitor.visit(self, origin)

# Monkey-patch if needed:
for cls in [Counterbalance, NoRepeat, StartWith, AbsoluteRank, SetRank, SetPosition, InnerBlock, OuterBlock]:
    cls.accept = accept

def copy_nested_constraints(design1, design2):
    visitor = ConstraintCopyVisitor(design1, design2)

    for constraint in design1.constraints:
        constraint.accept(visitor, origin='design1')

    for constraint in design2.constraints:
        constraint.accept(visitor, origin='design2')

    return visitor.constraints



def can_nest(d1, d2):
    if isinstance(d1, Plans) and isinstance(d2, Plans):
        return True
    
class RandomDesignError(Exception):
    pass

def handle_empty_design(des):
    des = copy.copy(des)
    if des.is_empty:
        var = ExperimentVariable("base", 1)
        des._add_variable(var)
        des.groups = Groups(1)
        des.constraints.append(Counterbalance(var, width = des.trials, height = 1, stride = [1,1]))
    return des

def nest(*, outer, inner):
    """
    Nest two designs to create a combined experimental design.

    Args:
        outer: The outer design object.
        inner: The inner design object.

    Returns:
        A new Design object representing the nested experimental design.
    """
    # Handle empty designs if necessary
    outer = handle_empty_design(outer)
    inner = handle_empty_design(inner)

    # Convert to RandomPlan if needed
    if not isinstance(inner, Replications) and (inner.has_random_variable() or inner.is_random()):
        inner = RandomPlan(inner.variables).num_trials(inner.get_width())
    if not isinstance(outer, Replications) and (outer.has_random_variable() or outer.is_random()):
        outer = RandomPlan(outer.variables)

    assert can_nest(outer, inner), "Designs cannot be nested due to incompatible structure."

    # Evaluate groups if necessary
    eval([inner, outer])
    num_outer_groups = len(outer.groups)
    if num_outer_groups == 0:
        outer.eval()
        num_outer_groups = len(outer.plans)

    total_groups = len(inner.groups) * num_outer_groups
    total_conditions = inner.get_width() * outer.get_width()

    combined_variables = combine_lists(inner.variables, outer.variables)
    combined_design = (
        Design()
        .limit_groups(total_groups)
        .num_trials(total_conditions)
    )

    # FIXME: The API should expose a way to add multifactor variables cleanly
    combined_design._add_variable(multifact(combined_variables))

    # NOTE: ugly way of handling this
    combined_design.random_var.extend(inner.random_var)
    combined_design.random_var.extend(outer.random_var)

    # Nest structural constraints and copy semantic ones
    combined_design.constraints.extend(nest_structure(inner, outer))

    copied_constraints = copy_nested_constraints(inner, outer)
    if copied_constraints:
        combined_design.constraints.extend(copied_constraints)

    return combined_design





def cross_structure(d1, d2):
    constraints = []
    # Match all variables from the outer design within each block matrix
    
    for i in range(len(d2.variables)):
        constraints.append(InnerBlock(
            d2.variables[i],
            1,
            len(d1.groups),
            stride = [1, 1]
        ))

     # Match all variables from the inner design across every block

    for i in range(len(d1.variables)):
        constraints.append(OuterBlock(
            d1.variables[i],
            d1.get_width(),
            len(d1.groups),
            stride = [1, 1]
        ))

 
    return constraints


# NOTE: this is an absolute mess :(
# FIXME: come back to this. Won't generalize...
# need to fix how the match blocks work, but this is not a priority
def cross(design1, design2):
    """
    Nest two designs together to create a combined experimental design.
    
    Args:
        design1: First design object
        design2: Second design object
        
    Returns:
        Combined design object
    """

    # first, check if one design is random
    # if so, conver to a special replications type.
    if (not isinstance(design1, Replications)) and (design1.has_random_variable() or design1.is_random()):
        design1 = RandomPlan(design1.variables).num_trials(design1.get_width())

    if (not isinstance(design2, Replications)) and (design2.has_random_variable() or design2.is_random()):
        design2 = RandomPlan(design2.variables)

    eval([design1, design2])

    # Calculate the total number of groups in the combined design
    total_groups = len(design1.groups) * len(design2.groups)
    # Combine variables from both designs
    combined_variables = combine_lists(design1.variables, design2.variables)
    
    # Calculate width1 (the product of all variable lengths in design1)
    width1 = design1.get_width()
    width2 = design2.get_width()
    
    assert width1 == width2
    total_conditions = width2 

    
    # Create a new design with the combined variables
    combined_design = ( Design()
                       .within_subjects(multifact(combined_variables))
                       .limit_groups(total_groups)
                       .num_trials(total_conditions)
                    )
   
   
    # need to modify the inner constraints region
    # Add counterbalance constraints from design1
    for constraint in design1.constraints:
        if isinstance(constraint, Counterbalance):
            if constraint.width and constraint.height:
                combined_design.constraints.append(
                    Counterbalance(
                        constraint.variable,
                        width=constraint.width,
                        height=constraint.height,
                        stride=constraint.stride
                    )
                )
            else: 
                combined_design.constraints.append(
                    Counterbalance(
                        constraint.variable,
                        width=width1,
                        height=len(design1.groups),
                        stride=constraint.stride
                    )
                )

        
        elif isinstance(constraint, NoRepeat):
            combined_design.constraints.append(
                NoRepeat(
                    constraint.variable,
                    constraint.width,
                    constraint.stride
                )
            )

        elif isinstance(constraint, AbsoluteRank):
                combined_design.constraints.append(
                    copy.copy(constraint)
                )

        elif isinstance(constraint, InnerBlock):
            combined_design.constraints.append(
                InnerBlock(constraint.variable, constraint.width, constraint.height, stride = [1, 1])
            )

        # # here I need to multiply stride by the number of conditions of the block variable
        elif isinstance(constraint, OuterBlock):
            combined_design.constraints.append(
                OuterBlock(constraint.variable, constraint.width, constraint.height, stride = [1, 1])
            )

    
    # need to modify out constraint region
    for constraint in design2.constraints:
        if isinstance(constraint, Counterbalance):
            stride_height = constraint.height if constraint.height else len(design1.groups)
            
            # Add counterbalance constraint for design2 variables
            combined_design.constraints.append(
                Counterbalance(
                    constraint.variable,
                    width=total_conditions,
                    height=total_groups,
                    stride=[stride_height, 1]
                )
            )

        elif isinstance(constraint, StartWith):
                combined_design.constraints.append(
                    StartWith(constraint.variable, constraint.condition)
                )
        
        elif isinstance(constraint, AbsoluteRank):
                combined_design.constraints.append(
                    copy.copy(constraint)
                )
        
        elif isinstance(constraint, NoRepeat):
        
              combined_design.constraints.append(
                NoRepeat(
                    constraint.variable,
                    constraint.width,
                    constraint.stride
                )
            )
    
        elif isinstance(constraint, InnerBlock):
            stride_height = constraint.height if constraint.height else len(design1.groups)
            combined_design.constraints.append(
                InnerBlock(constraint.variable, constraint.width,constraint.height*len(design1.groups), stride = [1, 1])
            )

        # # here I need to multiply stride by the number of conditions of the block variable
        elif isinstance(constraint, OuterBlock):
            stride_height = constraint.height if constraint.height else len(design1.groups)
            combined_design.constraints.append(
                OuterBlock(constraint.variable, constraint.width, constraint.height*len(design1.groups), stride = [stride_height, 1])
            )

    combined_design.constraints.extend(cross_structure(design1, design2))
    combined_design.random_var.extend(design1.random_var)
    combined_design.random_var.extend(design2.random_var)
    
    return combined_design


from z3 import *
from .candl import *

class BitVectors:
    def __init__(self, n, variables):
        self.variables = variables
        self.len = self.determine_num_bits()
        self.z3_representation =  [BitVec(f"C_{index + 1}", self.len) for index in range(n)]

    def get_possible_values(self):
        values = []
        k = 0
        for i in range(len(self.variables)):
            variable = self.variables[i]
            n = variable.n

            if not i:
                values.extend([x for x in range(n)])
            else:
                k += int(math.ceil(math.log(self.variables[i-1].n, 2))) + 1

                new_values = []
                for v in values:
                    for y in range(n):
                        x = y << k 
             
                        new_values.append(x + v)
                
                values.extend(new_values)

            if not len(values):
                values.extend([x for x in range(n)])

        return set(values)

    def determine_num_bits(self):
        bitvec_length = 0
        for variable in self.variables:
            bits = int(math.ceil(math.log(variable.n, 2)))
            bitvec_length += bits

        # necessary for two's compliment bit
        bitvec_length += len(self.variables)
        return bitvec_length
    

    def get_lower_bits(self, var):
        var_index = self.variables.index(var)
        bit_index = 0
        # for all the variables already represented
        for i in range(var_index):
            # length of all of the other variable bit representations
            bit_index += get_num_bits(len(self.variables[i]))

        # accounts for two's compliment bit
        bit_index += var_index
        return bit_index
    
    def get_variables(self):
        return self.z3_representation
    
    def get_variable_assignment(self, var, z3):
        lo = self.get_lower_bits(var)
        length = get_num_bits(len(var))
        return Extract(lo+length, lo, z3)
    
    def get_variable_assignments(self, var, z3s):
        
        assignments = []

        for z3 in z3s:
            assignments.append(self.get_variable_assignment(var, z3))

        return assignments




import itertools
from .variable import ExperimentVariable
from.condition import ExperimentCondition
import numpy as np
from itertools import product
import math
from z3 import *
import random

def distinct_or(variables):
    ret = []
    for i in range(len(variables)):
        for j in range(i + 1, len(variables)):
            temp = []
            for k in range(len(variables[i])):
                temp.append(variables[i][k] != variables[j][k])
            ret.append(Or(temp))

    return ret


def generate_conditions(participants, variable, n):
    """
    Generates a randomized sequence of conditions for each participant.
    
    :param participants: Number of participants
    :param conditions: List of conditions
    :param trials_per_condition: Number of times each condition should appear
    :return: Dictionary mapping participant IDs to their condition sequences
    """
    
    experiment_data = []
    if variable: 
        conditions = variable.conditions
        for _ in range(participants):
            experiment_data.append(random.sample(conditions, n))
   
    return experiment_data


# cartesian product of variable conditions to form experimental conditions
def conditions_from_vars(*argv):

    for arg in argv:
        assert isinstance(arg, ExperimentVariable)

    variable_conditions = [arg.conditions for arg in argv]
    conditions = list(itertools.product(*variable_conditions))

    conditions = [ExperimentCondition(f"condition{i+1}", *conditions[i]) for i in range(len(conditions))]



    return conditions


# move out of this class
def shape_array(arr, shape):
    return np.array(arr).reshape(shape).tolist()

def flatten_array(arr):
    return np.array(arr).flatten().tolist()

# FIXME
def create_indexing(dim, dims):
    dim_indexings = []
    # combination of all levels of each dimension, except for the one 
    # we want to put a constraint on
    keys = list(product(*[set(range(dims[i])) for i in range(len(dims)) if i != dim]))

    # this whole thing is a paradigm that I should be able
    # to use for various tasks
    for tup in keys:
        indexing = []
        count = 0
        for index in range(len(dims)):
            if index == dim:
                indexing.append(slice(None))
            else:
                indexing.append(tup[count])
                count += 1
     
        dim_indexings.append(indexing)

    return dim_indexings


def create_indexing_2(dim, dims):
    dim_indexings = []

    # this whole thing is a paradigm that I should be able
    # to use for various tasks
    for val in range(dims[dim]):
        indexing = []
        for index in range(len(dims)):
            if index == dim:
                indexing.append(val)
            else:
                indexing.append(slice(None))
     
        dim_indexings.append(indexing)

    return dim_indexings



def get_elements_of_dim(arr, shape, indexing):

    return np.array(arr).reshape(shape)[*indexing].tolist()

def all_elements_of_dim(dim, arr, shape):

    dim_indexings = create_indexing(dim, shape)


    dim_variables = []
    for indexing in dim_indexings:

        dim_variables.append(get_elements_of_dim(arr, shape, indexing))


    return dim_variables

def get_num_bits(n):
    return int(math.ceil(math.log(n, 2)))

def all_ones(n):
    return (1 << n) - 1


# FIXME: incorrect processing if there is a factor. See how majority does this. We probably want to decouple 
def get_dim_variables(arr, shape, dim, factor = None, level = None):
    if factor != None:
        dim_indexings = create_indexing_2(factor, shape)
        indexing = dim_indexings[level]

        arr_to_constrain = get_elements_of_dim(arr, shape, indexing)
        arr_to_constrain = flatten_array(arr_to_constrain)
       

        shape = tuple([shape[i] for i in range(len(shape)) if i != factor])
     
        dim_variables = all_elements_of_dim(dim - 1, arr_to_constrain, shape)

    else:
        dim_variables = all_elements_of_dim(dim, arr, shape)

    return dim_variables




def combine_lists(l1, l2):
    combined_variables = l1.copy()
    combined_variables.extend(l2)

    return combined_variables

def create_directory_for_file(filepath):
    directory = os.path.dirname(filepath)
    if directory:
        os.makedirs(directory, exist_ok=True)


def as_list(variables):
    if isinstance(variables, list):
        return variables
    else:
        return [variables]
    


def count_values(d: dict) -> dict:
    """Count the frequency of values in a dictionary."""
    counts = {}
    for value in d.values():
        counts[value] = counts.get(value, 0) + 1
    return counts


def factorial_product_of_counts(counts: dict) -> int:
    """Return the product of factorials of all counts."""
    return math.prod(math.factorial(count) for count in counts.values())

def calculate_plan_multiplier(num_conditions: int, num_vars: int, num_trials: int) -> float:
    """Compute the multiplier for total_n_plans based on design width and condition count."""
    if num_conditions == 1:
        return 1
    if num_trials > num_conditions:
        num_repeats = int(num_trials / num_conditions)
        denom = math.prod(math.factorial(num_repeats) for _ in range(num_vars))
        return math.factorial(num_conditions) / denom
    elif num_trials < num_conditions:
        return math.factorial(num_conditions) / math.factorial(num_conditions - num_trials)
    else:
        return math.factorial(num_conditions)



from z3 import *


class Constraint:
    def __init__(self, variable):
        self.variable = variable

    def get_variable(self):
        return self.variable

class BlockConstraint(Constraint):
    def __init__(self, variable, width, height, stride):
        super().__init__(variable)
        self.width = width
        self.height = height
        self.stride = stride

class PlanConstraint(Constraint):
    def __init__(self, variable, width, stride):
        super().__init__(variable)
        self.stride = stride
        self.width = width


class InnerBlock(BlockConstraint):
    """Represents a block structure in the experimental design."""
    def __init__(self, variable, width, height, stride= [1,1]):
        super().__init__(variable, width, height, stride)

class OuterBlock(BlockConstraint):
    """Represents a block structure in the experimental design."""
    def __init__(self, variable, width, height, stride):
        super().__init__(variable, width, height, stride)

    def __str__(self):
        return f"{self.variable}: \n\t{self.width}\n\t{self.height}\n"

class NoRepeat(PlanConstraint):
    def __init__(self, variable, width= None, stride= 1):
        if width is None:
            width = len(variable)
        super().__init__(variable, width, stride)

   
class Counterbalance(BlockConstraint):
    """Defines counterbalancing for variables."""
    def __init__(self, variable, width=0, height=0, stride=(1, 1)):
        super().__init__(variable, width, height, stride)
 
    def __str__(self):
        return f"COUNTERBALANCE: {self.width, self.height, self.stride}"
    
class StartWith(Constraint):
    def __init__(self, variable, condition):
        super().__init__(variable)
        # FIXME: modifying variables to contain list
        self.condition = condition

class SetRank(Constraint):
    def __init__(self, variable, condition, rank, condition2):
        super().__init__(variable)
        self.condition = condition
        self.rank = rank
        self.condition2 = condition2 

    
class AbsoluteRank(Constraint):
    def __init__(self, variable, condition, rank):
        super().__init__(variable)
        self.ranks = dict.fromkeys(variable.conditions, 0)
        self.add_rank(condition, rank)

    def add_rank(self, condition, rank):
        self.ranks[condition] = rank

    

class SetPosition(Constraint):
    def __init__(self, variable, condition, pos):
        super().__init__(variable)
        self.condition = condition
        self.position = pos




 


# I think I can get rid of this class

# Condition represents one possible "total condition" 
# pass in metaconditions to construct Condition, 
# along with constraints on the condition
class ExperimentCondition():
    # argv consists of Variable types
    def __init__(self, label):
        self.label = label


    





from z3 import *

from .blocks import BlockFactor

class Sequence(BlockFactor):
    """The PossibleOrders class represents all of the possible
        orders of conditions in an experiment, where each order
        is an experimental group

    n: the number of conditions in each order
    we use argv so the user can pass any number of variables
    because experiments have varying amount of variables
    """
    def __init__(self, n, *argv):

        BlockFactor.__init__(self, [i for i in range(n)])

       


    
    







    



# External libraries
from z3 import *             # Requires: pip install z3-solver
import math
# Planet core modules (installed via `pip install -e .`)
from planet.variable import ExperimentVariable, MultiFactVariable, multifact
from planet.design import Design, Plans
from planet.unit import Groups
from planet.constraint import (
    Counterbalance
)


class Replications(Plans):
    """Main class for creating experimental designs."""
    def __init__(self, n):
        super().__init__()
        var = ExperimentVariable("replications", 1)
        self._add_variable(var)
        self.groups = Groups(1)
        self.num_trials(n)
        self.constraints.append(Counterbalance(var, width = n, height = 1, stride = [1,1]))

    def num_trials(self, n):
        self.trials = n
        return self
    
    def _add_variable(self, variable):
        assert isinstance(variable, ExperimentVariable)

        if isinstance(variable, MultiFactVariable):
            variables = variable.variables
        else:
            variables = [variable]

        self.variables.extend(variables)
        self.ws_variables.extend(variables)
    
    
    def get_width(self):
        return self.trials or math.prod(v.n for v in self.ws_variables)
    


# External dependencies
from z3 import *  # requires `pip install z3-solver`
import math
import pandas as pd
from planet.logger import logger
# Internal modules (installed via `pip install -e .`)
from planet.unit import Groups
from planet.orders import Sequence
from planet.variable import MultiFactVariable, multifact
from planet.constraint import (
    StartWith, Counterbalance, NoRepeat,
    InnerBlock, OuterBlock, Constraint,
    SetRank, SetPosition, AbsoluteRank
)
from planet.designer import Designer
from planet.randomizer import Randomizer
from planet.formatter import LatexExport
from planet.candl import *
from planet.candl import generate_conditions
from planet.helpers import *
from planet.narray import *



class DesignWarning(UserWarning):
    pass

class DesignError(Exception):
    """Custom exception for errors in the Design class."""
    pass

class DesignVariable:
    def __init__(self, variable):
        self.variable = variable
        self.constraint_spec = {
            "Counterbalance":None,
            "NoRepeat":None,
            "AbsoluteRank":None,
            "OuterBlock":None,
            "InnerBlock":None,
        }

    def add_constraint(self, constraint):
        if self.constraint_spec[constraint.__name__] is None:
            self.constraint_spec[constraint.__name__] = constraint

class Plans:
    def __init__(self):
        self.sequence = None
        self.variables = []
        self.ws_variables = []
        self.bs_variables = []
        self.groups = None
        self.constraints = []
        self.trials = 0
        self.designer = Designer()
        self.plans = None
        # test
        self.random_var = []
        self.previous_snapshot = None
        
        self.design_variables = dict()

    def is_random(self):
        return not self.counterbalanced and not self.is_constrained
    
    def _add_design_variable(self, variable):
        if variable in self.design_variables:
            self.design_variables[variable] = DesignVariable(variable)

        
     
    
    #FIXME: come back to this
    def _add_variable(self, variable):
        assert isinstance(variable, ExperimentVariable)

        self._add_design_variable(variable)

        if isinstance(variable, MultiFactVariable):
            variables = variable.variables
        else:
            variables = [variable]

        self.variables.extend(variables)
        self.ws_variables.extend(variables)

class RandomPlan(Plans):
    """This is like creating a vector of random variables."""
    def __init__(self, variables):
        super().__init__()
        for var in variables:
            self._add_variable(var)
            self.random_var = [var]
            
        self.groups = Groups(1)
        self.constraints.append(NoRepeat(self.random_var[0], self.get_width()))

    def num_trials(self, n):
        self.trials = n
        return self
    
    
    
    def get_width(self):
        return self.trials or math.prod(v.n for v in self.ws_variables)

class Design(Plans):
    """Main class for creating experimental designs."""
    def __init__(self):
        super().__init__()
        self.rank_constraints = {}

        # FIXME
        # var = ExperimentVariable("base", 1)
        # self._add_variable(var)
        # self.constraints.append(Counterbalance(var, width = 0, height = 1, stride = [1,1]))
  
    def to_latex(self):
        # FIXME: won't work with random plans
        matrix = self.get_plans()
        formatter = LatexExport(matrix)
        formatter.to_latex()

    def num_trials(self, n):
        self.trials = n
        return self
    
    def _add_variable(self, variable):
        assert isinstance(variable, ExperimentVariable)

        if isinstance(variable, MultiFactVariable):
            variables = variable.variables
        else:
            variables = [variable]

        self.variables.extend(variables)
        self.ws_variables.extend(variables)
    
    def add_variable(self, variable, l):
        assert isinstance(variable, ExperimentVariable)

        # FIXME
        self._add_design_variable(variable)

        if isinstance(variable, MultiFactVariable):
            variables = variable.variables
        else:
            variables = [variable]

        self.variables.extend(variables)
        l.extend(variables)
    
    def between_subjects(self, variable):
        self.add_variable(variable, self.bs_variables)
        return self
    
    

    def within_subjects(self, variable):
        self.add_variable(variable, self.ws_variables)
        width = self.get_width()
        # FIXME: should probably decouble this constraint block concept?
        constraint = NoRepeat(variable, width=width)
        # self.design_variables[variable].add_constraint(constraint)
        self.constraints.append(constraint)
        return self
    
    def limit_groups(self, n):
        self.groups = Groups(n)
        return self
    
    def has_random_variable(self):
        return bool(self.random_var)
    
    
    def _determine_random_width(self, rand):
        width = self.get_width()
        div = 1
        for constraint in self.constraints:
            if isinstance(constraint, OuterBlock) and constraint.variable == rand:
                width = constraint.width
  
            elif isinstance(constraint, InnerBlock) and constraint.variable == rand:
                div = constraint.width
        
        return int(width/div)
    

    def _determine_random_span(self, rand):
        span = 1
        for constraint in self.constraints:
            if isinstance(constraint, InnerBlock) and constraint.variable == rand:
               span = constraint.width
        return span
    
    def _instantiate_random_variables(self, n, rand):
        # NOTE to self: this will only work if there is one random variable :)
        """
        Think about this like instantiating the elements of a matrix of random variables
        """
        assert self.plans is not None
        random_index = self.variables.index(rand)
        width = self._determine_random_width(rand)
        span = self._determine_random_span(rand)
        # randomly generates plans for every block of random var. 
        # rand_vars = self._generate_random_variables(int(n*self.get_width()/width/span), self.random_var, width) 
        # self.apply_randomization(rand_vars, width, span, random_index, n)
        randomizer = Randomizer(rand, width, span, random_index, int(n*self.get_width()/width/span), n, self.plans)
        self.plans = randomizer.get_plans()
    
    @property
    def is_modified(self):
        return self.snapshot() != self.previous_snapshot

    def snapshot(self):
        return str(hash(tuple(self.constraints))) + str(self.get_width()) + "_" + str(len(self.groups))

    def _eval_plans(self, n = None):
        self.identify_random_vars()

        if self.is_empty:
            return []
        elif self.is_random():
            self.random_var.extend(self.variables)
        if self.plans is not None and not self.is_modified:
            return self.plans
        else:
            self.eval()
            self.previous_snapshot = self.snapshot()
   
        if self.has_random_variable():
            assert n is not None
            n = math.ceil(n/len(self.plans)) * len(self.plans)
            for rand in self.random_var:
                self._instantiate_random_variables(n, rand)

    

    def check_plans(self, plans):
        if self.plans.size <= 0:
            raise DesignError("The constraints specified on this design are not compatible")
    
    def get_plans(self, n = None):
        self._eval_plans(n)

        try:
            self.check_plans(self.plans)
        except DesignError as e:
            logger.warning("Design results in no viable plans. Consider a different experiment.")
            

        
        return self.plans
    
    
    def get_width(self):
        return self.trials or math.prod(v.n for v in self.ws_variables)
    
    def start_with(self, variable, condition):
        assert isinstance(variable, ExperimentVariable)

        condition = as_list(condition)
        rank = 1
        for c in condition:
            self.absolute_rank(variable, c, rank)
        rank+=1
  
        return self
    
    def set_position(self, variable, condition, pos):
        assert isinstance(variable, ExperimentVariable)
        self.constraints.append(SetPosition(variable, condition, pos))
        return self
    
    def set_rank(self, variable, condition, rank, condition2):
        assert isinstance(variable, ExperimentVariable)
        self.constraints.append(SetRank(variable, condition, rank, condition2))
        return self
    
    def absolute_rank(self, variable, condition, rank):
        assert isinstance(variable, ExperimentVariable)
        if variable not in self.rank_constraints:
            self.rank_constraints[variable] = len(self.constraints)
            self.constraints.append(AbsoluteRank(variable, condition, rank))
        else:
            pos = self.rank_constraints[variable]
            rank_constraint = self.constraints[pos]
            rank_constraint.add_rank(condition, rank)

        return self
    
    def extract_counterbalance_group(self, constraint):
        """Extract variables and condition count from a Counterbalance constraint."""
        var = constraint.get_variable()
        num_conditions = len(var)
        if isinstance(var, MultiFactVariable):
            return (var.get_variables(), num_conditions)
        return ([var], num_conditions)
    
    def _determine_num_plans(self):
  
        """Determine the number of experimental plans based on constraints and trial width."""
        if self.groups:
            return self.groups  # Assume already set

        counterbalanced_groups = []
        contains_ranked_var = False
        total_n_plans = 1

        for constraint in self.constraints:
            if isinstance(constraint, Counterbalance):
                group = self.extract_counterbalance_group(constraint)
                counterbalanced_groups.append(group)

            elif isinstance(constraint, AbsoluteRank):
                contains_ranked_var = True
                rank_counts = count_values(constraint.ranks)
                total_n_plans *= factorial_product_of_counts(rank_counts)

        for variables, num_conditions in counterbalanced_groups:
            num_trials = self.get_width()
            total_n_plans *= calculate_plan_multiplier(num_conditions, len(variables), num_trials)

        self.groups = Groups(int(total_n_plans) if (counterbalanced_groups or contains_ranked_var) else 0)
    
                


    def _eval(self):
        # Get the width of the design
        width = self.get_width()
        sequence = Sequence(width)
        self._determine_num_plans()
        

        self.designer.start(
            self.groups, 
            sequence, 
            self.variables
        )

        # Handle between-subjects constraints
        for variable in self.bs_variables:
            self.constraints.append(
                InnerBlock(
                    variable,
                    width,
                    1
                )
            )
        
        # NOTE: ensure no downstream effects :)
        # self.designer.solver.all_different()
        self.designer.eval_constraints(self.constraints, self.groups, width)

    def test_eval(self):
        self._eval()
        return self.designer.eval_all()
        

    def eval(self):
        assert not self.is_empty
        self._eval()
        plans = self.designer.eval()
        self.plans = plans

    @property
    def counterbalanced(self):
        # FIXME: not handled for replicated trials
        return any(isinstance(c, Counterbalance) or isinstance(c, AbsoluteRank) for c in self.constraints)
     
    # NOTE: need to make this much faster. Hashmap with var -> constraints?
    def identify_random_vars(self):
        for v in self.variables:
            if not any((isinstance(c, Counterbalance) or isinstance(c, AbsoluteRank)) and c.variable == v for c in self.constraints):
                self.random_var.append(v)

    @property
    def is_constrained(self):
        # FIXME: not handled for replicated trials
        return any(isinstance(c, OuterBlock) or isinstance(c, InnerBlock) for c in self.constraints)
    

    @property
    def is_empty(self):
        # FIXME: not handled for replicated trials
        return not self.variables 

        
    def counterbalance(self, variable, w = 0, h = 0, stride = [1, 1]):
        assert isinstance(variable, ExperimentVariable)

        self.constraints.append(Counterbalance(variable, width = w, height = h, stride = stride))
        return self
    
    def _new_ws_variables(self):
        if self.ws_variables:
            return self.ws_variables[0] if len(self.ws_variables) == 1 else multifact(self.ws_variables)
        else:
            return None

    def _new_bs_variables(self): 
        if self.bs_variables:   
            return self.bs_variables[0] if len(self.bs_variables) == 1 else multifact(self.bs_variables)
        else:
            return None
        
    def _combined_conditions(self, bs_conditions, ws_conditions):
        conditions = []
        if self.ws_variables and self.bs_variables:
            for i in range(len(bs_conditions)):
                conditions.append([bs_conditions[i][0] + "-" + ws_conditions[i][j] for j in range(len(ws_conditions[0]))])
        elif self.bs_variables:
            conditions = bs_conditions
        elif self.ws_variables:
            conditions = ws_conditions
        else:
            raise RuntimeError("No variables defined to generate conditions.")
        
        return conditions
        
    def random(self, n):
        assert len(self.ws_variables) or len(self.bs_variables) 

        ws_variable = self._new_ws_variables()
        ws_conditions = generate_conditions(n, ws_variable,self.get_width())

        
        bs_variable = self._new_bs_variables()
        bs_conditions = generate_conditions(n, bs_variable,1)
            
        return self._combined_conditions(bs_conditions, ws_conditions)
    




from planet.candl import generate_conditions
from planet.constraint import InnerBlock, OuterBlock



class Randomizer:
    def __init__(self, variable, width, span, random_index, n, participants, plans):
        rand_vars = self._generate_random_variables(n, variable, width)
        self.randomized_plans = self._apply_randomization(rand_vars, width, span, random_index, participants, plans)
      
    def _generate_random_variables(self, n, variable, width):
        ws_conditions = generate_conditions(n, variable, width)
        return ws_conditions
    
    def get_blocks(self, plan, width, span):
        """Chunk the plan into blocks of size width * span."""
        block_size = width * span
        return [plan[i:i + block_size] for i in range(0, len(plan), block_size)]

    def replace_condition(self, condition_str, replacement, index_to_replace):
        """Replace part of a condition string at the given index."""
        parts = condition_str.split("-")
        parts[index_to_replace] = replacement
        return "-".join(parts)

    def _apply_randomization(self, rand_vars, width, span, random_index, n, plans):
        new_plans = []

      
        for plan_idx, plan in enumerate(plans):
            blocks = self.get_blocks(plan, width, span)
            reps_per_plan = int(n / len(plans))

            for rep_idx in range(reps_per_plan):
                new_plan = []

                for block_idx, block in enumerate(blocks):
                    # FIXME: does  not work when random variables are the outer design
                    rand_idx = rep_idx * len(blocks) + plan_idx * int(n/len(plans)) + block_idx
                    rand_var = rand_vars[rand_idx]
                    for cond_idx in range(width):
                        for within_block_idx in range(span):
                            index = cond_idx * span + within_block_idx
                            old_condition = block[index]
                            replacement = rand_var[cond_idx]
                            new_condition = self.replace_condition(old_condition, replacement, random_index)
                            new_plan.append(new_condition)

                new_plans.append(new_plan)

        return new_plans
    
    def get_plans(self):
        return self.randomized_plans




  


# External libraries
from z3 import *                # Requires: pip install z3-solver
import numpy as np

# Core system modules (planet package)
from planet.orders import Sequence
from planet.candl import *          #
from planet.helpers import *       
from planet.narray import *
from planet.solver import BitVecSolver
from planet.constraint import (
    StartWith, Counterbalance, NoRepeat,
    InnerBlock, OuterBlock, Constraint,
    SetRank, SetPosition, AbsoluteRank
)
from planet.variable import MultiFactVariable
from planet.unit import Units

# NOTE: combine this with des pls
class Designer:
    """The assignment class matches every unit to an order of conditions
        based on constraints set on unit variables, such as block factors
        of participants, pid, and the state of observation of a subject. 

    """
    def __init__(self):
        self.constraints = []

    def start(self, subjects, sequence, variables=[]):
        assert isinstance(sequence, Sequence)
        assert isinstance(subjects, Units)
        assert len(variables) > 0

        self.units = subjects
        self.seq = sequence
        self.num_trials = len(sequence)
        self.variables = variables
        self.shape = self.determine_shape()
        self.solver = BitVecSolver(self.shape, self.variables)

    def _get_experiment_variables(self, constraint):
        if isinstance(constraint.variable, MultiFactVariable):
            return constraint.get_variable().get_variables()
        else:    
            return [constraint.get_variable()]

    def eval_constraints(self, constraints, groups, width):
        # Process constraints
        for constraint in constraints:
            assert isinstance(constraint, Constraint)
            variable = self._get_experiment_variables(constraint)

            match constraint:
                
                case Counterbalance():
                    # FIXME
                    if groups:
                        constraint.width = (
                            constraint.width if constraint.width else width
                        )
                        constraint.height = (
                            constraint.height if constraint.height else len(groups)
                        )

                    self.counterbalance(
                        variable,
                        w=constraint.width,
                        h=constraint.height,
                        stride=constraint.stride,
                    )

                case NoRepeat():
            
                        self.solver.all_different(
                            variable, 
                            constraint.width, 
                            constraint.stride
                        )

                case StartWith():
                    self.start_with(
                        constraint.variable, 
                        constraint.condition
                    )
                
                case SetPosition():
                    self.set_position(
                        constraint.variable, 
                        constraint.condition,
                        constraint.position
                    )

                case SetRank():
                    self.set_rank(
                        constraint.variable, 
                        constraint.condition,
                        constraint.rank,
                        constraint.condition2
                    )

                case AbsoluteRank():
                    self.absolute_rank(
                        constraint.variable, 
                        constraint.ranks
                    )

                case InnerBlock():
                    self.match_inner(
                        constraint.variable, 
                        constraint.width, 
                        constraint.height
                    )

                case OuterBlock():
                    self.match_outer(
                        constraint.variable, 
                        constraint.width, 
                        constraint.height
                    )
                case _:
                    raise ValueError("Could not identify proper constraint")
        
    def determine_shape(self):
        if len(self.units): 
            n = len(self.units)
        else: 
            n = 1

        return tuple([n, self.num_trials])
    
    # FIXME: creating block matrix for specific test case 
    # Note: use for creating blocks

    # NEED TO DECOUPLE THIS
    def match_inner(self, variable, width, height):
        # get number of block matrices per column
        n = int(self.shape[0] / height)
        # get number of block matrices per row
        m = int(self.shape[1] / width)
       
        for i in range(n):
            for j in range(m):
                self.solver.match_block(
                    variable, 
                    [
                        (i*height + 0, i * height  + height, 1)
                        , (j*width + 0, j * width + width, 1)
                    ]
                )

        



    def match_outer(self, v, w, h):
        # NEEDS TO BE it's own func / constraint option. Don't treat these together 
        for i in range(h):
            for j in range(w):
                
                self.solver.match_block(v, [(i, self.shape[0], h), (j, self.shape[1], w)])

    def counterbalance(self, v, w, h, stride = [1, 1]):
        block = [(0, h, stride[0]), (0, w, stride[1])]
        self.solver.counterbalance(block, v)

    def start_with(self, variable, condition):
        self.solver.start_with(variable, variable.conditions.index(condition))

    def set_rank(self, variable, condition, rank, condition2):
        self.solver.set_rank(variable, variable.conditions.index(condition), rank, variable.conditions.index(condition2))
    
    def set_position(self, variable, condition, pos):
        self.solver.set_position(variable, variable.conditions.index(condition), pos)
    
    def absolute_rank(self, variable, ranks):
        transformed_ranks = {variable.conditions.index(condition): rank for condition, rank in ranks.items()}
        self.solver.absolute_rank(variable, transformed_ranks)
    

    # def get_groups(self, model):
    #     if not len(self.units):
    #         self.units.n = len(model)



    # NOTE: this is with a bitvec representation...
    # ensure that this works
    def eval(self):
        # perhaps this is where we create a different class?
        # we have so many new instance vars
        # these should maybe be classes or something? 
        if len(self.units):
            model = self.solver.get_one_model()
            return self.decode(model)
        else:
            model = self.solver.get_all_models()
            # self.get_groups(model)
            # FIXME
            return np.array(self.solver.encoding_to_name(model, self.variables))
    
    def decode(self, model):
        if not len(model):
            return np.array([])
        else:
            reshaped_model = np.array(model).reshape(self.shape).tolist()
            return np.array(self.solver.encoding_to_name(reshaped_model, self.variables))
  
    # NOTE: this is with a bitvec representation...
    # ensure that this works
    def eval_all(self):
        # perhaps this is where we create a different class?
        # we have so many new instance vars
        # these should maybe be classes or something? 
        if len(self.units):
            models = self.solver.get_all_models()

            ret = []
            for model in models:
                assert len(model) > 0
                model = np.array(model).reshape(self.shape).tolist()
                ret.append(np.array(self.solver.encoding_to_name(model, self.variables)))

            return ret
        else:
            model = self.solver.get_all_models()
            self.get_groups(model)
            return np.array(self.solver.encoding_to_name(model, self.variables))
            



from .blocks import BlockFactor


# inherit from variable? 
class Groups(BlockFactor):
    """ a Unit represents a singular unit that participates in a study. Units
            get randomly assigned to an experimental group during assignment

    n: id of the unit
    """
    def __init__(self, n, labels = None):
        BlockFactor.__init__(self, [i for i in range(n)])
        self.n = n
        self.labels = labels

    def expand_groups(self, num_groups):
        assert num_groups % self.n == 0
        return Groups(self.n)



from z3 import *
from .helpers import *
from .orders import Sequence
from .narray import *
import numpy as np
from .candl import *
from .solver import  BitVecSolver
from .unit import Units, Groups, Clusters
import pandas as pd
import math
import time

import duckdb

def determine_plans(units, design):
  
    return design.get_plans(n=len(units))

class Assignment:
    def __init__(self, units, plans):
        assign_counterbalance(units, plans)
        self.units = units
        self.plans = plans
    
    def format_plans(self):
        matrix = determine_plans(self.units, self.plans)
        ret = ""

        for i in range(len(matrix)):
            ret += f"plan {i+1}:\n\t" 
            for j in range(len(matrix[i])):

                end = "" if j == len(matrix[i]) - 1 else "\t" 
                conditions = matrix[i][j].split("-")
                test = [f"{self.plans.variables[x].name} = {conditions[x]}" for x in range(len(conditions))]
                conditions = ", ".join(test)

                ret += f"trial {j+1}: {str(conditions)}\n" + end
        return ret
    
   
    def to_csv(self):
        plans = self.plans.eval()
        trials = [f"trial{i+1}" for i in range(len(plans[0]))]
        df = pd.DataFrame(plans, columns=trials)
        units_df = self.format_assignment()
        df = df.reset_index()
        df = df.rename(columns={'index': 'row_number'})
        df = units_df.merge(df, how='inner', left_on='plan', right_on='row_number')
        df = df.drop("row_number", axis=1)
        df.to_csv('../outputs/assignment.csv', index=False)
    
    def format_assignment(self):
        return duckdb.sql(f"select * from {self.units.table}").to_df()
        
    def __str__(self):
        ret = f"""***EXPERIMENT PLANS***\n\n{str(self.format_plans())} \n\n***ASSIGNMENT***\n\n{str(self.format_assignment())}"""
        return ret

def assign_subunits(units, parent):
    duckdb.sql(f""" update {units.table}
                        set plan = subunit_assignment.plan
                        from 
                            (
                                select units.pid id, expanded_cluster_assignment.plan plan  
                                from {units.table} units, 
                                    ( select plan, unnest(subunits) test 
                                      from {parent.table} 
                                    ) as expanded_cluster_assignment
                                where units.pid = expanded_cluster_assignment.test
                            ) subunit_assignment

                        where subunit_assignment.id = {units.table}.pid
        
           """)


def construct_assignment_table(table, units, num_plans, num_participants):
     # Create a temporary table to store plan assignments
    duckdb.sql("CREATE TABLE members (plan INT)")

   
    required_participants = math.ceil(num_participants/num_plans) * num_plans if num_participants else num_plans
    for i in range(num_participants, required_participants):
        duckdb.sql(f"insert into {units.table} values ({i+1}, 0)")
    
    num_per_group = required_participants // num_plans  # Number of participants per plan
    
    # Insert plan assignments into the temporary table
    for i in range(num_plans):
        for _ in range(num_per_group):
            duckdb.sql(f"INSERT INTO members VALUES ({i})")
    
    # Randomly distribute the plan assignments to the participants
    duckdb.sql(f"""
        UPDATE {table}
        SET plan = assignment.plan
        FROM (
            SELECT {table}.pid AS id, members.plan
            FROM (
                SELECT *, ROW_NUMBER() OVER (ORDER BY UUID()) AS rand
                FROM members
            ) members
            JOIN {table} ON {table}.pid = members.rand
        ) AS assignment
        WHERE assignment.id = {table}.pid
    """)
    duckdb.sql("DROP TABLE members")

    for i in range(num_participants, required_participants):
        duckdb.sql(f"update {units.table} set pid = -1 where pid > {num_participants}")



def assign_units(units, plans):
    """
    Assigns participants (units) to different plans in a balanced manner.
    Ensures that the number of participants is evenly distributed across all plans.

    Parameters:
    units : object
        An object containing participant data and metadata.
    plans : object
        An object containing plan assignment logic.
    """
    
    # Evaluate the units object to ensure it's up-to-date
    table = units.get_table()
    plans = determine_plans(units, plans)
    num_plans = len(plans)  # Total number of available plans
    num_participants = len(units)  # Total number of participants

    if num_plans:
        construct_assignment_table(table, units, num_plans, num_participants)
    
   
def assign_counterbalance(units, plans, parent = None):
    if parent is not None:
        assign_subunits(units, parent)
    else:
        assign_units(units, plans)
        
    # NOTE: assign subunits to the same plan as their parent unit was assigned to 
    if isinstance(units, Clusters):
        assign_counterbalance(units.units, plans, parent = units)


def assign(units, plans):
    return Assignment(units, plans)


       


from z3 import *
from .blocks import BlockFactor
import duckdb
import time

class Unit:
    """ a Unit represents a singular unit that participates in a study. Units
            get randomly assigned to an experimental group during assignment

    n: id of the unit
    """
    def __init__(self, n):
        self.n = n

class Units:
    def __init__(self, n, table = "participants"):
        self.n = n
        self.attributes = []
        self.hash = str(time.time()).replace(".", "")
        self.table = table + self.hash
        self.evaled = False

    def add_attribute(self, attr):

        assert isinstance(attr, BlockFactor)
        self.attributes.append(attr)

    
    def eval(self):
        if not self.evaled:
            duckdb.sql(f"create table {self.table} ( pid int, plan int )")
            for i in range(self.n):
                duckdb.sql(f"insert into {self.table} values ({i+1}, 0)")
            self.evaled = True

    def get_table(self):
        self.eval()
        return self.table

    def get_plan(self, n):
        return duckdb.execute(f"select plan from {self.table} where pid = {n}").fetchone()[0]

    def get_attributes(self):
        return self.attributes
    
    def __len__(self):
        return self.n 
    

class Clusters(Units):
    """A Cluster represents a group of Units that are grouped together for a specific purpose."""
    def __init__(self, units, n):
        assert units.n % n == 0
        super().__init__(n, "clusters")

        assert units is not None
        self.units = units

    def get_units(self, n):
        ret = duckdb.sql(f"select subunits from {self.table} where pid = {n}").fetchnumpy()["subunits"][0]
        return list(ret)

    def eval(self):
        self.units.eval()
        super().eval()

        duckdb.sql(f"""ALTER TABLE {self.table}
                        ADD subunits INTEGER[]""")
        
     
     
        limit = int(self.units.n/self.n)
        for i in range(1, self.n+1):   
            
            duckdb.sql(f"""update {self.table} set subunits = (select array_agg(units.pid) from 
                                    (select      *, 
                                                row_number() over(order by uuid()) rand
                                    from        {self.units.table} where pid not in (select unnest(subunits) from {self.table} ) limit {limit}) units)
                                    
                                    where pid = {i}""")
            



       

# inherit from variable? 
class Participants(Units):
    """ a Unit represents a singular unit that participates in a study. Units
            get randomly assigned to an experimental group during assignment

    n: id of the unit
    """
    def __init__(self, n=0):
        Units.__init__(self, n)

   

# inherit from variable? 
class Groups(Units):
    """ a Unit represents a singular unit that participates in a study. Units
            get randomly assigned to an experimental group during assignment

    n: id of the unit
    """
    def __init__(self, n):
        Units.__init__(self, n)
        self.groups = [Group(i) for i in range(n)]

    def expand_groups(self, num_groups):
  
        assert num_groups % self.n == 0
        
        new_groups = Groups(num_groups)
        for attr in self.attributes:

            new_groups.add_attribute(attr)
        return new_groups
    


class Group:
    def __init__(self, n):
        self.n = n
        self.labels = []

    def add_label(self, label):
        self.labels.append(label)

    



        
    



import logging

print("logger")
class ColorFormatter(logging.Formatter):
    BRIGHT_PINK = "\033[95m"  # bright magenta (pink)
    WHITE = "\033[97m"
    RESET = "\033[0m"

    def format(self, record):
        # Format the original message first
        msg = super().format(record)

        # Replace the levelname (e.g., WARNING) with colored version
        # record.levelname is typically uppercase, e.g., 'WARNING'
        colored_levelname = f"{self.BRIGHT_PINK}{record.levelname}{self.RESET}"

        # Replace only the first occurrence of the levelname in the message
        # (usually appears at the start)
        msg = msg.replace(record.levelname, colored_levelname, 1)

        # Wrap the rest of the message in white color
        # (Assuming formatter outputs levelname followed by message)
        # Since we replaced levelname already, ensure msg ends with reset
        return f"{self.WHITE}{msg}{self.RESET}"


logger = logging.getLogger("planet")
if not logging.getLogger().hasHandlers():
    logging.basicConfig(level=logging.ERROR)
    logger.setLevel(logging.WARNING)
    handler = logging.StreamHandler()
    formatter = ColorFormatter('%(levelname)s: %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    

logger.propagate = False




from z3 import *
import numpy as np
from .candl import *
from .helpers import *
from functools import reduce
from .bitvector import BitVectors
import itertools
import warnings

# NOTE: to decouple, should we have an ndarray class that 
# has z3 vars as special type? not sure if this infrastructure is necessary?

class BitVecSolver:
    def __init__(self, shape, variables):
        
        # how can we use translate without circular dependencies? 
        # some of these vars are related and should be stored together as a struct 
        self.shape = shape
        self.variables = variables
        self.num_z3_vars = np.prod(self.shape)
        self.solver = z3.Optimize()
        self.bitvectors = []
       
        self.create_z3_variables()
        self.constrain_z3_values()
        self.distinguish_rows()

        self.rank_functions = {}
        self.ranks = {}
 

    # you can come up with a better name
    def constrain_z3_values(self):
        """ for all of the z3 variables relating to a specific variable, 
         ensure that it can only be assigned to one of the levels
         of the specific variable """
        lower_bit = 0
   
        for variable in self.variables:
            # NOTE: z3 vectors is stored as a flattened array
            for index in range(self.num_z3_vars):
                # NOTE: range(lo, hi) should give all levels of variable as int
                lo = 0
                hi = len(variable) - 1
                # required number of bits to encode a given value
                required_bits = int(math.ceil(math.log(variable.n, 2))) 
                self.solver.add(
                        And(
                            lo <= Extract(
                                    lower_bit+required_bits, 
                                    lower_bit, 
                                    self.z3_conditions[index]
                                ), 
                            hi >= Extract(
                                lower_bit+required_bits, 
                                lower_bit, 
                                self.z3_conditions[index]
                            )
                        )  
                )

            # gives us index of the lower bit of the next variable
            lower_bit += required_bits + 1

        
    # FIXME: function that merges these
    # z3 variable for the overal condition assigned to a unit. 
    # for example, an n by n lating square has n*n z3 variables 
    # for each cell of the square. represented by the flattened array
    def create_z3_variables(self):
        self.bitvectors = BitVectors(self.num_z3_vars, self.variables)
        self.z3_conditions = self.bitvectors.get_variables()
    
    def block_columns(self, arr, width, stride=1):
        return np.array(arr)[:, 0:width:stride]

    def all_different(self, v=None, width = None, stride = 1):
        if v is None: 
            return
        # could you prettify this?
        dim_variables = get_dim_variables(self.z3_conditions, self.shape, 1)

        if width is not None:
            dim_variables = list(self.block_columns(dim_variables, width, stride))

        test = lambda x: [self.bitvectors.get_variable_assignment(z, x) for z in v]

        for arr in dim_variables:
            # FIXME (fixed but I'm not convinced)
            if v is not None:
                self.solver.add(distinct_or(list(map(test, arr))))
            else:
                self.solver.add(Distinct(arr))

    def n_trials(self):
        return self.shape[1]

    # NOTE: no rows can repeat in a given matrix 
    def distinguish_rows(self):
        plans = get_dim_variables(self.z3_conditions, self.shape, 1)
        for i in range(len(plans)):
            for j in range(len(plans)):
                if i != j:
                    assignment = []
                    for n in range(len(plans[j])):
                        assignment.append(plans[i][n] != plans[j][n])
                    self.solver.add(Or(assignment))



    def establish_ranking(self, var):
        orders = get_dim_variables(self.z3_conditions, self.shape, 1)
        n = self.n_trials()

        for var in self.rank_functions:
            rank = self.rank_functions[var]
            for order in orders:
                self.solver.add([rank(order[i]) >= rank(order[i+1]) for i in range(n-1)])



    def count(self, l, x, f):
        test = []
        for e in l:
            test.append(If(f(e, x), 1, 0))
        return sum(test)
    

    def block_array(self, arr, block = []):
        return np.array(arr)[block[0][0]:block[0][1]:block[0][2], block[1][0]:block[1][1]:block[1][2]]
    
    def start_with(self, variable, condition):
        plans = np.array(get_dim_variables(self.z3_conditions, self.shape, 1))

        arr = plans[:, 0]
        for z3 in arr:
            self.solver.add(self.bitvectors.get_variable_assignment(variable, z3) == condition)


    def set_position(self, variable, condition, pos):
        plans = np.array(get_dim_variables(self.z3_conditions, self.shape, 1))

        arr = plans[:, pos]
        for z3 in arr:
            self.solver.add(self.bitvectors.get_variable_assignment(variable, z3) == condition)


    def set_rank(self, var, condition, rank, condition2):
        x = BitVec("x", self.bitvectors.len)
        y = BitVec("y", self.bitvectors.len)

        if var not in self.rank_functions:
            self.rank_functions[var] = Function(f'rank_{str(var)}', BitVecSort(self.bitvectors.determine_num_bits()), IntSort())
            self.establish_ranking(var)

        rank = self.rank_functions[var]
        self.solver.add(
            ForAll([x,y],
                Implies(
                    And(self.bitvectors.get_variable_assignment(var, x) == condition, 
                        self.bitvectors.get_variable_assignment(var, y) == condition2),
                    rank(x) > rank(y)
                )
            )
        )

    def absolute_rank(self, var, ranks):
        x = BitVec("x", self.bitvectors.len)

        orders = get_dim_variables(self.z3_conditions, self.shape, 1)
        n = self.n_trials()

        rank = Function(f'rank_{str(var)}', BitVecSort(self.bitvectors.determine_num_bits()), IntSort())
        for order in orders:
            self.solver.add([rank(order[i]) >= rank(order[i+1]) for i in range(n-1)])

        for condition in ranks:
            self.solver.add(
                ForAll([x],
                    Implies(
                        self.bitvectors.get_variable_assignment(var, x) == condition,
                        rank(x) == ranks[condition]
                    )
                )
            )
            

    def counterbalance(self, block=[], variables=None):
        """
        Apply counterbalancing to ensure equal occurrence of variable combinations.
        
        Args:
            block: List of blocks to apply (default empty list)
            variables: The variables to counterbalance (required)
            
        Returns:
            None: Modifies the solver in-place
        """
      
        # Get dimensional variables from z3 conditions
        plans = get_dim_variables(self.z3_conditions, self.shape, 1)
        

        # Apply blocks if provided
        if len(block):
            plans = self.block_array(plans, block)
        

        # Transpose the plans for processing
        plans = np.transpose(plans)

    

        # Initialize counts for tracking variable combinations
        counts = []
        
        # Define helper functions for value processing
        def get_possible_values(variable):
            """Return set of possible values for a variable (0 to len-1)"""
            return set(range(len(variable)))
        
        def test_assignment(variable, position):
            """Get variable assignment at the given position"""
            return self.bitvectors.get_variable_assignment(variable, position)
        
        # Generate all possible combinations of variable values
        value_combinations = list(itertools.product(
            *[get_possible_values(variable) for variable in variables]
        ))
        
        # Process each plan

        for plan_idx in range(len(plans)):
            # Map variables to their assignments in the current plan
            assignments = list(zip(*[
                list(map(lambda p: test_assignment(variable, p), plans[plan_idx])) 
                for variable in variables
            ]))
            
            # Count occurrences of each value combination
            for value_combo in value_combinations:
                # Define equality constraint function
                def check_equality(x, expected):
                    """Check if x equals expected for all indices"""
                    return And([expected[i] == x[i] for i in range(len(x))])
                
                # Count occurrences and store
                counts.append(self.count(assignments, value_combo, check_equality))
        
        # Add constraints to ensure equal counts for all combinations
        for i in range(len(counts)):
            for j in range(i+1, len(counts)):
                self.solver.add(counts[i] == counts[j])
                
    def match_block(self, variable, block = []):
        plans = get_dim_variables(self.z3_conditions, self.shape, 1)
        block = self.block_array(plans, block)

        flat_block = np.array(block).flatten()
        for i in range(len(flat_block)):
            for j in range(i, len(flat_block)):
                a1 = self.bitvectors.get_variable_assignment(variable, flat_block[i])
                a2 = self.bitvectors.get_variable_assignment(variable, flat_block[j])
                self.solver.add(a1==a2)

        # works when z3 representation is a matrix 
    def get_one_model(self):
        all_assignments = []
    
        if (self.solver.check() == sat):
            all_assignments = []
            model = self.solver.model()
      
            for var in self.z3_conditions:
                all_assignments.append(model.evaluate(model[var]))
 

        
        return all_assignments
    
    # works when z3 rep iterates through many arrays
    # NOTE: should I merge these two functions?
    def get_all_models(self):
        all_orders = []
        self.solver.push()
        count = 0
        capacity = 10000
        while self.solver.check() == sat and count < capacity:
            model = self.solver.model()
            block = []
            order = []
            for var in self.z3_conditions:
                block.append(var != model[var])
                order.append(model.evaluate(model[var]))

            all_orders.append(order)
            self.solver.add(Or(block))
            count+=1
        if self.solver.check() == sat:
            warnings.warn(f"There are more orders not captured in the model due order capacity (currently set to {capacity})")

        self.solver.pop()
        return all_orders


    def encoding_to_name(self, all_orders, variables):
        # this makes life much easier
        z3_assignments = flatten_array(all_orders)
        num_vars = len(variables)
        
        orders = []
        for z3_assignment in z3_assignments:
            # this is an integer value, represented as a binary string
            assignment = int(str(z3_assignment))
            # this is where we will store the string rep of a condition
            # which is a concatination of it's variable assignments
            decoded_assignment = ""
 
            # # there is probably a more efficient way to do this
            for i in range(num_vars):
                # get the ith variable's assignment
                variable_assignment = assignment 
                # shift right variable assignment unitl desired var assignment 
                # is in the lower bits (+1 to account for two's compliment)
         
                for j in range(i):
                    variable_assignment >>= get_num_bits(len(variables[j])) + 1

                # get lower bits (associated with given variable)
                mask =  all_ones(get_num_bits(len(variables[i]))+ 1)
                variable_assignment &= mask

                # FIXME
                if variable_assignment >= len(variables[i]):
                    condition = "dropout"
                else:
              
                    # FIXME: this is silly, but need to inverse order to match array indexing
                    # variable_assignment = len(variables[i]) - 1 - variable_assignment
                    condition = variables[i].conditions[variable_assignment]

                # conder variable assignment to a string representation 
                decoded_assignment+=str(condition)
                # FIXME; hyphenate output for readability
                if i < len(variables)-1:
                    decoded_assignment+="-"

            orders.append(decoded_assignment)

        # return back in original form
        return shape_array(orders, np.array(all_orders).shape)
    
    def name_to_encoding(self, model):

        # this makes life much easier
        model_assignments = flatten_array(model)
        num_vars = len(self.variables)
        
        for i in range(len(model_assignments)):
            assignment = model_assignments[i].split("-")

            # there is probably a more efficient way to do this
            for j in range(num_vars):
                # get the ith variable's assignment
                variable_assignment = assignment[j]
                condition = self.variables[j].condition_map[variable_assignment]
                __z3__ = self.bitvectors.get_variable_assignment(self.variables[j], self.z3_conditions[i])
                self.solver.add(__z3__ == condition)

    def check_model(self, model):
  
        self.solver.push()

        self.name_to_encoding(model, self.variables)
        result = self.solver.check()

        self.solver.pop()

        return result



import pandas as pd
from .candl import *

class LatexExport:

    def __init__(self, design_data):
        """
        design_data: Could be a pandas DataFrame or
        any structure that can be converted into one.
        """
        self.df = design_data

    def to_latex(self):
        trials = [f"trial{i+1}" for i in range(len(self.df[0]))]
        df = pd.DataFrame(self.df, columns=trials)
        try:
            filepath = "outputs/design.tex"
            create_directory_for_file(filepath)
            with open(filepath, 'w', encoding='utf-8') as tex_file:
                tex_file.write(df.to_latex())
            print(f"success")
        except Exception as e:
            print(f"An error occurred: {e}")


from itertools import product
from z3 import *

def add_dimension(dims, unit_variables):
    if len(unit_variables) == 1:
        return [f"unit_" for i in range(dims[0])]
    
    else:
        return [add_dimension(dims[1:], unit_variables[1:]) for i in range(dims[0])]
    
def get_narray_elem(indices, narray):
    curr = narray
    i = 0
    while type(curr) == list:
        curr = curr[indices[i]]
        i += 1

    return curr

def modify_narray_elem(indices, narray, f):
    curr = narray
    i = 0
    arr = None
    while type(curr) == list:
        arr = curr
        curr = curr[indices[i]]
        i += 1

    arr[indices[i-1]] =  f(indices[i-1], arr)

    return narray

def str_from_list(l):
    ret = ""
    for s in l:
        ret+=str(s+1)
    return ret

def construct_units(narray, dims):
    keys = list(product(set(range(dims[0])), set(range(dims[1]))))
  
    for key in keys:
        add = lambda i, arr : arr[i] + str_from_list(key)
        narray = modify_narray_elem(key, narray, add)

    return narray

def construct_z3(narray, dims):
    keys = list(product(set(range(dims[0])), set(range(dims[1]))))
    z3_fun = lambda i, arr : Int(arr[i]) 
    for key in keys:
        narray = modify_narray_elem(key, narray, z3_fun)

    return narray




from setuptools import setup, find_packages

requirements=[
    "bitarray==2.9.3",
    "bitstring==4.2.3",
    "contourpy==1.3.1",
    "cycler==0.12.1",
    "duckdb==1.1.3",
    "fonttools==4.55.3",
    "Jinja2==3.1.6",
    "kiwisolver==1.4.7",
    "lark==1.2.2",
    "MarkupSafe==3.0.2",
    "matplotlib==3.10.0",
    "numpy==2.1.2",
    "packaging==24.2",
    "pandas==2.2.3",
    "pillow==11.0.0",
    "pyparsing==3.2.0",
    "python-dateutil==2.9.0.post0",
    "pytz==2025.2",
    "six==1.17.0",
    "tzdata==2025.2",
    "z3-solver==4.13.3.0",
]

setup(
    name='planet-dsl',
    version='0.1.3',
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    author='London Bielicke',
    description='A DSL for defining experimental designs',
    install_requires = requirements
)




def DEBUG(string, color = "35"):
    print("\033[+" + color + "+m{}\033[00m" .format(string))

# this is a helper function
def add(list, exp, n):
    if n == 1:
        exp = list[0]
        return exp

    exp = list[n - 1] + list[n - 2]
    list[n - 2] = exp
    return add(list, exp, n-1)


from z3 import *
from .variable import ExperimentVariable

class BlockFactor(ExperimentVariable):
    def __init__(self, levels, name = ""):
        if name == "":
            name = self.__class__
        ExperimentVariable.__init__(self, name, num_options=len(levels), options = levels)
        self.n = len(levels)
        self.levels = levels
        self.sequence_length = self.n 
        self.solver = z3.Solver()
        self.constraints = []
        self.z3_vectors = None # rep each var in z3
        self.all_orders = [] # stores all possible orders given constraint

    def __len__(self):
        return self.n


class Block: 
    def __init__(self, variable, assignment):
        self.variable = variable
        self.designer = assignment
